 !-------------------------------------------
 !Module to calculate various spin related functions.
 !-------------------------------------------

 MODULE module_spin_functions

 USE, INTRINSIC :: iso_fortran_env

 IMPLICIT NONE
 
 !-----------------------------------------------------------------------------------
 !Public variables
 !-----------------------------------------------------------------------------------
 INTEGER,PARAMETER :: dpr = REAL64
 REAL(kind=dpr), PARAMETER :: pi_value = 3.14159265358979323846_dpr
 REAL(kind=dpr), ALLOCATABLE, DIMENSION(:) :: Xi_lookup,inverse_Xi_lookup
 INTEGER, PARAMETER :: num_lookuppoints = 10000
 REAL(kind=dpr), PARAMETER :: Xithreshold = 0.001_dpr 
 REAL(kind=dpr), PARAMETER :: Xizerotolerance = 0.000001_dpr
 
 CONTAINS

 SUBROUTINE generate_spin_lookup_tables()
 !===================================================================================
 INTEGER :: i   
 ALLOCATE(Xi_lookup(num_lookuppoints))
 ALLOCATE(inverse_Xi_lookup(num_lookuppoints))
 Xi_lookup=0.0_dpr
 inverse_Xi_lookup=0.0_dpr
 DO i=1,num_lookuppoints
   Xi_lookup(i) = calculate_Xi(real(i,dpr)/num_lookuppoints)
   inverse_Xi_lookup(i) = calculate_inverse_Xi(i*2.0_dpr*pi_value/num_lookuppoints)
 END DO
 
 END SUBROUTINE generate_spin_lookup_tables


 !-------------------------------------------
 !Calculating Xi.
 !-------------------------------------------
 ELEMENTAL FUNCTION calculate_Xi(tau) RESULT(Xi)
 REAL(kind=dpr), INTENT(IN) :: tau
 REAL(kind=dpr) :: Xi, Xi_derivative,inv_Xi, tau_estimate, Xi_estimate

 IF (tau < 0.0_dpr) THEN
   Xi = 0.0_dpr
 ELSE IF (tau > 1.0_dpr) THEN
   Xi=2.0_dpr*pi_value
 ELSE IF (tau < Xithreshold) THEN
   Xi=4.0_dpr*pi_value*tau/3.0_dpr
 ELSE IF(tau > (1.0_dpr - Xithreshold) ) THEN
   Xi=2.0_dpr*pi_value + (tau - 1.0_dpr)*49.956_dpr + (tau - 1.0_dpr)**2*8370.6_dpr
 ELSE
   Xi = pi_value*((1.0_dpr/(tau**2) - 1.0_dpr)*log((1.0_dpr - tau)/(1.0_dpr + tau)) + 2.0_dpr/tau)
 END IF

 RETURN 

 END FUNCTION calculate_Xi


 !-------------------------------------------
 !Calculating Xi_derivative.
 !-------------------------------------------
 ELEMENTAL FUNCTION calculate_Xi_derivative(tau) RESULT(Xi_derivative)
 REAL(kind=dpr), INTENT(IN) :: tau
 REAL(kind=dpr) ::  Xi, Xi_derivative,inv_Xi, tau_estimate, Xi_estimate

 IF ( tau < Xithreshold ) THEN
   Xi_derivative = 4.0_dpr*pi_value/3.0_dpr
 ELSE IF ( tau > ( 1.0_dpr - Xithreshold )) THEN
   Xi_derivative=50.0_dpr
 ELSE
   Xi_derivative=2.0_dpr * pi_value * (log (( 1.0_dpr +tau ) / ( 1.0_dpr -tau)) - 2.0_dpr*tau)/(tau**3)
 END IF

 END FUNCTION calculate_Xi_derivative


 !-------------------------------------------
 !Calculating inverse_Xi.
 !-------------------------------------------
 ELEMENTAL FUNCTION calculate_inverse_Xi(Xi_value) RESULT(inv_Xi) 
 REAL(kind=dpr), INTENT(IN) :: Xi_value
 REAL(kind=dpr) ::  Xi_derivative,inv_Xi, tau_estimate, Xi_estimate
 INTEGER:: i

 IF (Xi_value < 0.0_dpr) THEN
   inv_Xi=0.0_dpr
 ELSE IF (Xi_value < Xizerotolerance) THEN
   inv_Xi=3.0_dpr * Xi_value/(4.0_dpr *pi_value)
 ELSE IF (Xi_value > 2.0_dpr*pi_value - Xizerotolerance) THEN
   inv_Xi=1.0_dpr
 ELSE
   tau_estimate= (Xi_value-0.036835_dpr*Xi_value*Xi_value )/( 4.178319_dpr-0.136129_dpr*Xi_value+0.038148_dpr*Xi_value*Xi_value)
   Xi_derivative = calculate_Xi_derivative(tau_estimate)
     do i=1,5
        Xi_estimate=calculate_Xi(tau_estimate)
        tau_estimate=tau_estimate+(Xi_value-Xi_estimate)/Xi_derivative
     END DO
     inv_Xi=tau_estimate
 END IF

 END FUNCTION calculate_inverse_Xi
 


 PURE FUNCTION fast_calculate_Xi(tau,Xi_lookup)
 !-----------------------------------------------------------------------------------
 INTEGER :: i  
 REAL(kind=dpr),INTENT(IN) :: Xi_lookup(num_lookuppoints)
 REAL(kind=dpr),INTENT(IN) :: tau
 REAL(kind=dpr) :: fast_calculate_Xi,threshold,f

 threshold = 1.0_dpr/num_lookuppoints
 IF (tau <= 0.0_dpr) THEN
   fast_calculate_Xi = 0.0_dpr
 ELSE IF (tau >= 1.0_dpr) THEN
   fast_calculate_Xi = 2.0_dpr*pi_value
 ELSE IF (tau < threshold) THEN
   fast_calculate_Xi = 4.0_dpr*pi_value*tau/3.0_dpr
 ELSE  
    i = floor(tau*num_lookuppoints)
    f = tau*num_lookuppoints - i
    fast_calculate_Xi = (1.0_dpr-f)*Xi_lookup(i) + f*Xi_lookup(i+1)
 END IF

 END FUNCTION fast_calculate_Xi


 PURE FUNCTION fast_calculate_inverse_Xi(Xi_value,inverse_Xi_lookup)
 !-----------------------------------------------------------------------------------
 REAL(kind=dpr),INTENT(IN) :: Xi_value,inverse_Xi_lookup(num_lookuppoints)
 REAL(kind=dpr) :: fast_calculate_inverse_Xi,threshold,scaled_Xi_value,f
 INTEGER :: i
 
 threshold = 1.0_dpr/num_lookuppoints
 scaled_Xi_value = Xi_value/(2.0_dpr*pi_value)
 IF (scaled_Xi_value <= 0.0_dpr) THEN
    fast_calculate_inverse_Xi = 0.0_dpr
 ELSE IF (scaled_Xi_value >= 1.0_dpr) THEN
    fast_calculate_inverse_Xi = 1.0_dpr
 ELSE IF (scaled_Xi_value < threshold) THEN
    fast_calculate_inverse_Xi = Xi_value*0.75_dpr/pi_value
 ELSE  
    i = floor(scaled_Xi_value*num_lookuppoints)
    f = scaled_Xi_value*num_lookuppoints - i
    fast_calculate_inverse_Xi = (1.0_dpr-f)*inverse_Xi_lookup(i) + f*inverse_Xi_lookup(i+1)
 END IF

 END FUNCTION fast_calculate_inverse_Xi 
 
 

 PURE FUNCTION calculate_theta_scalar(a,b_projection,Xi_lookup)
 !-----------------------------------------------------------------------------------
 REAL(kind=dpr),INTENT(IN) :: a,b_projection,Xi_lookup(num_lookuppoints)
 REAL(kind=dpr) :: calculate_theta_scalar,tau,Xi
 REAL(kind=dpr), PARAMETER :: local_tolerance = 1.0e-10_dpr 
 
 IF ((abs(b_projection) < local_tolerance) .or. (a < local_tolerance)) THEN
    calculate_theta_scalar = 0.0_dpr
 ELSE
    tau = (abs(b_projection)/a)
    Xi = fast_calculate_Xi(tau,Xi_lookup)
    calculate_theta_scalar = b_projection*Xi/(2.0_dpr*abs(b_projection))
 END IF
 
 END FUNCTION calculate_theta_scalar


 PURE FUNCTION calculate_theta_vector(a,b_vector,Xi_lookup)
 !-----------------------------------------------------------------------------------
 REAL(kind=dpr),INTENT(IN) :: a,b_vector(3),Xi_lookup(num_lookuppoints)
 REAL(kind=dpr) :: calculate_theta_vector(3),mag_b_vector,tau,Xi,b(3)
 REAL(kind=dpr), PARAMETER :: local_tolerance = 1.0e-10_dpr 
 
 mag_b_vector = sqrt(dot_product(b_vector,b_vector))
 IF ((mag_b_vector > local_tolerance) .and. (a > local_tolerance)) THEN
   tau = (mag_b_vector/a)
   Xi = fast_calculate_Xi(tau,Xi_lookup)
   calculate_theta_vector = b_vector*Xi/(2.0_dpr*mag_b_vector)
 ELSE
   calculate_theta_vector = 0.0_dpr
 END IF   

 END FUNCTION calculate_theta_vector
 

 END MODULE module_spin_functions
